/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)talkd.c	5.4 (Berkeley) 6/18/88";
#endif /* not lint */

/*
 * The top level of the daemon, the format is heavily borrowed
 * from rwhod.c. Basically: find out who and where you are; 
 * disconnect all descriptors and ttys, and then endless
 * loop on waiting for and processing requests
 *
 * -----------------------------------------------------------------------------
 *
 * Modified by Paul B. Ricks (Kent State University) 2/26/92 to support xcomm.
 *
 * -----------------------------------------------------------------------------
 */


#include <syslog.h>
#include <arpa/inet.h>
#include "talkd.h"

#ifndef TRUE
#define	TRUE 	1
#define FALSE	0
#endif

#define WAITING_TIME 5

CTL_MSG		request;
int	msglen = sizeof(request);
CTL_RESPONSE	response;

int	sockt;
int	debug = 0;
int	timeout();
long	lastmsgtime;

int	server_ok;
char	hostname[32];
struct  hostent *hp;
int	needto_reply = TRUE;
struct  sockaddr_in fromaddr;
struct	sockaddr *fsp = (struct sockaddr *)&fromaddr;
int	fromlen = sizeof(fromaddr);
struct 	sockaddr_in server_addr = { AF_INET };
struct 	sockaddr *gsp = (struct sockaddr *)&server_addr;
int	gsplen = sizeof(*gsp); 	

#ifdef XCOMM_NO_SU
void syslog();
#endif

#define TIMEOUT 30
#define MAXIDLE 120
#define	PING_SECS 20

void	do_nothing(), sigalrm_handler();

main(argc, argv)
	int argc;
	char *argv[];
{
	int cc;
	FILE *fp;
	struct servent *sp;
	struct itimerval itimer;
	register CTL_MSG *mp = &request;


#ifndef XCOMM_NO_SU

	if (getuid()) {
	    fprintf(stderr, "%s: getuid: not super-user\n", argv[0]);
	    exit(1);
	}

	openlog("xcommd", LOG_PID, LOG_DAEMON);
#endif
	
	if (gethostname(hostname, sizeof(hostname) - 1) < 0) {
	    syslog(LOG_ERR, "gethostname: %m");
	    _exit(1);
	}
        if ((hp = gethostbyname(hostname)) == NULL) {
            fprintf(stderr, "Couldn't get host machine address.\n");
            exit(-1);
        }

#ifndef XCOMM_NO_SU

	if (chdir("/dev") < 0) {
	    syslog(LOG_ERR, "chdir: /dev: %m");
	    _exit(1);
	}
#endif

	parseArgs(argc, argv);

	if (server_ok) {
	    if ((sp = getservbyname("xcommads", "udp")) != (struct servent *)0) 
	        server_addr.sin_port = htons(sp->s_port);
 	    else
	        server_addr.sin_port = htons(XCOMM_ADS_PORT);
	}

	if (server_ok) {
	    int psockt;
	    struct sockaddr_in junk;
	    struct sockaddr *jp = (struct sockaddr *)&junk;
	    int junklen = sizeof(junk);

	    server_ok = FALSE;
	    signal(SIGALRM, (void (*)())do_nothing);

	    mp->type = PING;
	    server_addr.sin_family = AF_INET;
	    mp->ctl_addr.sa_family = AF_INET;
	    psockt = get_socket(SOCK_DGRAM, 0);
	    if (psockt != -1) {

		strncpy(mp->l_name, hostname, NAME_SIZE - 1);
		strncpy(mp->r_name, hostname, NAME_SIZE - 1);
		mp->l_name[NAME_SIZE -1] = mp->r_name[NAME_SIZE -1] = '\0';

		itimer.it_interval.tv_sec = WAITING_TIME;
		itimer.it_interval.tv_usec = 0;
		itimer.it_value = itimer.it_interval;
		setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);

		if ((cc = sendto(psockt, (char *)mp, sizeof(*mp), 0, gsp, gsplen)) > -1) {
		     if (recvfrom(psockt, (char *)mp, sizeof(*mp), 0, jp, &junklen) == sizeof(*mp))
			 server_ok = TRUE;
		}
		itimer.it_value.tv_sec = itimer.it_interval.tv_sec = 0;
		setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);
		close(psockt);
	    }
	}

#ifdef XCOMM_NO_SU

	if ((sockt = get_socket(SOCK_DGRAM, XCOMM_DPORT)) < 0) {
	    fprintf(stderr, "Error getting socket for daemon.\n");
	    exit(0);
	}
	dup2(sockt, 0);
	close(sockt);
#endif 

	fromlen = sizeof(*fsp);
	signal(SIGALRM, (void (*)())sigalrm_handler);
	for (;;) {
		extern int errno;
#ifdef XCOMM_NO_SU
		itimer.it_interval.tv_sec = ONE_MINUTE * 10;
#else
		itimer.it_interval.tv_sec = ONE_MINUTE;
#endif
		itimer.it_interval.tv_usec = 0;
		itimer.it_value = itimer.it_interval;
		setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);

		cc = recvfrom(0, (char *)mp, sizeof(*mp), 0, fsp, &fromlen);

		itimer.it_value.tv_sec = itimer.it_interval.tv_sec = 0;
		setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);

		if (cc != sizeof(*mp)) {
		    if (errno != EINTR)
			syslog(LOG_WARNING, "recvfrom: %m");
		    continue;
		}

		if (debug) {
                    fprintf(stderr, "\07\07\07Request from:\n");
                    fprintf(stderr, "Port: %d\n", fromaddr.sin_port);
                    fprintf(stderr, "Family: %d\n", fromaddr.sin_family);
                    fprintf(stderr, "Address: %-s\n", inet_ntoa(fromaddr.sin_addr.s_addr));
                    fprintf(stderr, "Requester: %-s\n", mp->l_name);
		    fprintf(stderr, "Request type: %d (%d)\n", mp->type, REGISTER);
		}

	        signal(SIGALRM, (void (*)())do_nothing);
		process_request(mp, &response);

		if (needto_reply) {

		    itimer.it_interval.tv_sec = WAITING_TIME;
		    itimer.it_interval.tv_usec = 0;
		    itimer.it_value = itimer.it_value;
		    setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);

		    cc = sendto(0, (char *)&response, sizeof(response), 0, fsp, fromlen);
		    if (cc != sizeof(response))
			syslog(LOG_WARNING, "sendto: %m");

		    itimer.it_value.tv_sec = itimer.it_interval.tv_sec = 0;
		    setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);
		}
		signal(SIGALRM, (void (*)())sigalrm_handler);
	}
}

void do_nothing()
{
   signal(SIGALRM, (void (*)())sigalrm_handler);
}

void sigalrm_handler()
{
   exit(0);
}

#ifdef XCOMM_NO_SU
void syslog()
{
	;
}
#endif
