#define notdef

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <netinet/in.h>

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)talkd.h	5.2 (Berkeley) 3/13/86
 *
 *
 *	Modified by Paul B. Ricks (Kent State University) to work with XComm.
 */

/*
 * This describes the protocol used by the talk server and clients.
 *
 * The talk server acts a repository of invitations, responding to
 * requests by clients wishing to rendezvous for the purpose of
 * holding a conversation.  In normal operation, a client, the caller,
 * initiates a rendezvous by sending a CTL_MSG to the server of
 * type LOOK_UP.  This causes the server to search its invitation
 * tables to check if an invitation currently exists for the caller
 * (to speak to the callee specified in the message).  If the lookup
 * fails, the caller then sends an ANNOUNCE message causing the server
 * to broadcast an announcement on the callee's login ports requesting
 * contact.  When the callee responds, the local server uses the
 * recorded invitation to respond with the appropriate rendezvous
 * address and the caller and callee client programs establish a
 * stream connection through which the conversation takes place.
 */

/*
 * Client->server request message format.
 */
typedef struct {
	u_char	vers;		/* protocol version */
	u_char	type;		/* request type, see below */
	u_char	answer;		/* not used */
	u_char	xcomm;		/* used to be a pad */
	u_long	id_num;		/* message id */
	struct	sockaddr addr;
	struct	sockaddr ctl_addr;
	long	pid;		/* caller's process id */
#define	NAME_SIZE	12
	char	l_name[NAME_SIZE];/* caller's name */
	char	r_name[NAME_SIZE];/* callee's name */
#define	TTY_SIZE	16
	char	r_tty[TTY_SIZE];/* callee's tty name */
} CTL_MSG;

/*
 * Server->client response message format.
 */
typedef struct {
	u_char	vers;		/* protocol version */
	u_char	type;		/* type of request message, see below */
	u_char	answer;		/* respose to request message, see below */
	u_char	xcomm;
	u_long	id_num;		/* message id */
	struct	sockaddr addr;	/* address for establishing conversation */
} CTL_RESPONSE;

#define	XCOMMFLAG	123

#define	TALK_VERSION	1		/* protocol version */

/* message type values */
#define LEAVE_INVITE	0	/* leave invitation with server */
#define LOOK_UP		1	/* check for invitation by callee */
#define DELETE		2	/* delete invitation by caller */
#define ANNOUNCE	3	/* announce invitation by caller */

/*
 * Operational parameters.
 */
#define MAX_LIFE	60	/* max time daemon saves invitations */
/* RING_WAIT should be 10's of seconds less than MAX_LIFE */
#define RING_WAIT	30	/* time to wait before resending invitation */


/*	xcomm flags 	
	-----------
*/

#define QUIT		6	/* use in terminating multiple xcomms */ 
#define REGISTER        10	/* register talk connection with master server */
#define RE_REGISTER	11	/* register talk connection with master server */
#define DE_REGISTER	12	/* de-register a connection with master server */
#define DE_REGISTERALL	13	/* de-register a connection with master server */
#define	PLEASE_RELAY	14	/* if ANNOUNCE, then relay message if possible */
#define	PING		15	/* used to determine whether the server is up */
#define PHONELINE	16	/* toggles xcomm on/off. */
#define	TALK_REQUEST	17

/* answer values */
#define SUCCESS		0	/* operation completed properly */
#define NOT_HERE	1	/* callee not logged in */
#define FAILED		2	/* operation failed for unexplained reason */
#define MACHINE_UNKNOWN	3	/* caller's machine name unknown */
#define PERMISSION_DENIED 4	/* callee's tty doesn't permit announce */
#define UNKNOWN_REQUEST	5	/* request has invalid type value */
#define	BADVERSION	6	/* request has invalid protocol version */
#define	BADADDR		7	/* request has invalid addr value */
#define	BADCTLADDR	8	/* request has invalid ctl_addr value */

#define	IS_HERE		30
#define	BAD_RELAYADDR	31
#define	COULDNT_RELAY	32

#define ON_LINE		2
#define OFF_LINE	3

#define VALID		1
#define	NOT_VALID	2

#define XCOMM_ADS_PORT	5008	
#define XCOMM_DPORT	5009	
#define ADDRESS_SERVER_HOSTNAME	"Chaos"
#define	SERVERWAITTIME	400

#define ONE_MINUTE	60
