/*
 * Copyright (c) 1992 Kent State University.
 * All rights reserved.  The Kent State University 
 * software License Agreement specifies the terms and conditions 
 * for redistribution.
 *
 *	@(#)xcommf0.c	1.1 (Kent State) 2/28/92
 *
 *	Author: Paul B. Ricks
 *
 *	The spirit of the Gnu Licensing Agreement Applies to this software and 
 *	Kent State University should be substituted where ever Gnu appears.
 *
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1992 Kent State University.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)xcommf0.c 1.1 (Kent State Univ.) 2/28/92";
#endif /* not lint */

/*

                            NO WARRANTY

   BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.

                     END OF TERMS AND CONDITIONS
*/
/*
 * This describes the functionality of the (Xcommunicate) xcomm version of the talk program.
 *
 *
 * The two files xcommf0.c and xcommf1.c contain routines that drive the X aspect
 * of the talk program.  The majority of the work is done by the child process. The
 * child process establishes connection with the callee and then passes the connected
 * stream socket to the parent.  The parent then handles the rest of the setting up.
 * There are two sockets used between the parent process and its child. One of these
 * socket is used for reporting status information and the other is used for transfering
 * "talk lines" from the child to the parent. The child basicly handles the networking
 * choir while the parent does the X'ing choir. When the call is being made, the parent
 * passes the string from the widget to the child. The child parses the string and passes
 * back relevant information about the callee and then proceeds to make the actual call.
 *
 * xcomm has the following features:
 *
 *	1.) if used with the xcomm adresss server, calls on the same local network do not
 *	    have to be qualified --- jdoe@kusaba.mcs.edu.  If jdoe happens to not be running
 *	    xcomm and is not logged on to the same machine, then address qualification must
 *	    be used otherwise xcomm behaves just like the regular talk program in this case.
 *	    Pressing <Return> initiates the call.
 *
 *	2.) The are 5 commands related to an in-coming call:
 *
 *		a) Accept --- accepts the call and pops up a window for that call.
 *		b) Refuse --- denies permission to connect
 *		c) Put on hold --- puts the call on hold.  This feature is activated
 *				   only if both sides are using xcomm. 
 *		d) Take message (answering machine) --- takes a message from the other side.
 *		e) Pick call --- takes a call off hold
 *
 *	3.) Record/Read message to retrieve messages left on the answering machine or
 *	    to leave messages for particular people. (goes by user names)
 *	    There is a limit on the length of both in-coming and out-going messages.
 *	    (1024 bytes - 1)
 *
 *	4.) Each user can have his/her own global message for everybody by using the
 *	    user name <anybody>. The order of acquiring messages proceeds from specific
 *	    user name, anybody, and then the build-in message.
 *
 *	5.) During a talk session, the panel of the other side can be recorded to a file.
 *	    after specifying the file, characters coming from the other side are then saved.
 *	    By pressing the <Record> button, one can toggle between recording and pausing.
 * 
 *	6.) Option menu that can put xcomm in answering machine mode without confirmation from
 *	    a user.  This menu also allows one to take his/her phone off line. The keys to
 *	    invoke this are <ctrl M>.
 *
 *
 *	7.) Supports for multiple line buffer pasting.
 *
 *	8.) call history logger
 * Note:
 *
 *	I didn't quite get the timing of displaying some of the various messages right.  This
 *	can easily be fixed if you one to. The routine that displays these messages take the
 *	following arguments:
 *
 *	   issueError(msg_string, flashing_count, frequency_index, bell_tone, ring_count).
 *
 *
 *	
 *	Sometimes, the timer might get out of whack, so it is possible for the status line
 *	to not be clear of old messages.  The old message will be cleared by the next
 *	call comming-in/going-out.
 */


#include	"xcommon.h"
#include	"xcomm.bmp"
#include	"ixcomm.bmp"
#include	"xcommLine.bmp"
#include	"ixcommLine.bmp"

unsigned long
msgIntervals[] = {
       Frequency_01, 
       Frequency_02, 
       Frequency_03, 
       Frequency_04, 
       Frequency_05, 
       Frequency_06,
       Frequency_07,
       Frequency_08
};


int xrelay_msg();
void CancelOpenFile(), StopRecording(), apoligize(), clear_textwin(), issueError(), ring_bell();
void TransferLine(), Quit(), PopupTalkPanel(), Beep(), Popdowntalkpanel(), MakeCall(), Beep(), msgRepeater();
void CloseTalkSocket(), ReadCharFromHim(), CloseTalkLine(), getTalkLine(), CloseRelayLine(), ProcessRelayMsg();
void SendCharToHim(), Copychar(), ReleaseString(), ReleaseInputRec(), SendKillLine(), SendDeleteChar(), SendDeleteWord();
void invitationRepeater(), issueRequest(), issueResponse(), SendCharsToHim();
void OkayOpenFile(), changeIcon(), hangUp(), RequestSelection(), getSelection();
void parseCMDLine(), OpenRecordingFile(), PopupFilepanel(), DialCallee();
void OpenRFile(); 

extern int checkForMessages();
extern void PopupHLog(), PopdownHLog(), PopupMessages(), PopupOptions(), PopdownOptions();
extern Widget create_optionpanel(), create_msgpanel();

static XtActionsRec actionLists[] = {
	{"beep", Beep},
	{"make-call", MakeCall},
	{"open-recording-file", OpenRFile},
	{"popup-options", PopupOptions},
	{"pop-options-down", PopdownOptions},
	{"transfer-line", TransferLine},
	{"send-line-kill", SendKillLine},
	{"send-delete-char", SendDeleteChar},
	{"send-delete-word", SendDeleteWord},
	{"send-his-char", SendCharToHim},
	{"request-selection", RequestSelection},
	{"popup-history-log", PopupHLog},
	{NULL, NULL}
};

static char *xcommVersion = " xcomm - version 1.11 ";
static char *DefaultSorryMsg = "\t(Message from XComm): Please try placing your call again\n\tin\
 a while.\0";
static Pixmap TalklinePixmap, icon_pixmap, iicon_pixmap, iTalklinePixmap;

Display *mydpy;
char *sorryMsg;
char testing = 0;
int tlogCounter = 0;
int logCounter = 0;
Bool ringBell = True;
XtTranslations trans;
char quitCode = QUIT_CODE;
String gbuf0, gsbuf0, gebuf0;
static int repeatCount;
Bool msgAnnouncer = False;
static XtAppContext xcomm;
static char panel_counter = 0;
static XtIntervalId requestId, callIndicatorID;
static XtInputId talkLineId, texceptionId, relayId, rexceptionId;
Widget create_talkpanel(), create_mainmenu();
Bool solidIcon, newPhonestate, oldPhonestate, newAAnswerstate, oldAAnswerstate;
Widget dial_w, typing_w, accept_w, status_w, refuse_w, callmenu_w, hangup_w, putonhold_w, 
       topLevel, takemsg_w, pickcall_w, options_w, message_w, retrievemsg_w, mainmenu_popup,
       hisLog_w, hisList_w, option_w, logcommand_w, logLabel_w, logQuit_w; 

typedef struct {
	int sockt, fd;
	char edchars[3];
	char recording;
	char *bp, *buf, *epos;
	Widget pw, w0, w1, w2, fw, rb, fb, fname_w;
	XtInputId inputId, exceptionId;
} sInputRec;


static struct {
	char id;
	Widget w;
	sInputRec *info;
} talkpanels[Max_Panels];

LOCAL_MSG pc_msg;
errorRec gerrordata;

static  char makecallTranslations[] =
	"<Key>Return: make_call() newline() beginning-of-line()\n\
	Ctrl<Key>H: delete-previous-character()\n\
	Ctrl<Key>W: delete-previous-word()\n\
	Ctrl<Key>U: beginning-of-line() kill-to-end-of-line()\n\
	<Key>Left: delete-previous-character()\n\
	Ctrl<Key>: beep()";

static  char mywinTranslations[] = 
	"Ctrl<Key>H: delete-previous-character()\n\
	Ctrl<Key>W: delete-previous-word()\n\
	Ctrl<Key>U: kill-to-end-of-line()\n\
	<Key>Left: delete-previous-character()\n\
	Ctrl<Key>: beep()\n\
	<Key>Return: transfer_line()";

static 	char defaultTranslations1[] =
	"<Key>: beep()\n <Btn1Down>: beep()\n\
	<Btn2Down>: beep()\n Ctrl<Key>: beep()";

static  char hiswinTranslations[] =
	"Ctrl<Key>a: delete-previous-character()\n\
	 Ctrl<Key>b: kill-to-end-of-line()\n\
       	 Ctrl<Key>c: delete-previous-word()";

char hisTtyBuf[LittleBuf], hisNameBuf[MAX_MSGBUF + 1], hisAddressBuf[MAX_MSGBUF + 1];

extern int callwaiting;
extern char my_edchars[];
extern onholdRec calls[];
extern int fds[], filed_con[];
extern int parentpid, childpid;
extern void hangUp(), xsigpipe_handler();

void do_xcommunicate(argc, argv)
	int	argc;
	char	**argv;
{
	register int i;

	argv[0] = xcommVersion;
    	topLevel = XtVaAppInitialize(&xcomm, "XComm", NULL, ZERO, &argc, argv, /*fallback_res*/NULL, NULL);
	mydpy = XtDisplay(topLevel);

	icon_pixmap = XCreateBitmapFromData(mydpy, RootWindowOfScreen(XtScreen(topLevel)),
					    (String)xcomm_bits, xcomm_width, xcomm_height);
	iicon_pixmap = XCreateBitmapFromData(mydpy, RootWindowOfScreen(XtScreen(topLevel)),
					    (String)ixcomm_bits, ixcomm_width, ixcomm_height);
	TalklinePixmap = XCreateBitmapFromData(mydpy, RootWindowOfScreen(XtScreen(topLevel)),
					    (String)xcomml_bits, xcomml_width, xcomml_height);
	iTalklinePixmap = XCreateBitmapFromData(mydpy, RootWindowOfScreen(XtScreen(topLevel)),
					    (String)ixcomml_bits, ixcomml_width, ixcomml_height);

	XtVaSetValues(topLevel, XtNiconPixmap, icon_pixmap, NULL);
    	XtAppAddActions(xcomm, actionLists, XtNumber(actionLists));

	for (i = 0; i < MAX_QUE_CALLS; i++)
	    calls[i].id = -1;

	newPhonestate = oldPhonestate = True;
	newAAnswerstate = oldAAnswerstate = False;

    	mainmenu_popup =  create_mainmenu(topLevel);
	options_w = create_optionpanel(topLevel);
	message_w = create_msgpanel(topLevel);

	for (i = 0; i < Max_Panels; i++) 
	     talkpanels[i].id = -1;

	relayId = XtAppAddInput(xcomm, fds[0], XtInputReadMask, ProcessRelayMsg, NULL);
	rexceptionId = XtAppAddInput(xcomm, fds[0], XtInputExceptMask, CloseRelayLine, NULL);

	sorryMsg = DefaultSorryMsg;
	(void)signal(SIGPIPE, (void (*)())xsigpipe_handler);	/* register this handler just in case */

    	XtRealizeWidget(topLevel);
    	XtAppMainLoop(xcomm);
}


Widget  create_mainmenu(toplevel)
	Widget	toplevel;
{
	String lbuf0;
    	Widget  a, l;
	int isthereNewMsg;
	Widget f, f2, f3, f4, p, b;

		            
        f = XtVaCreateManagedWidget("xcommf", formWidgetClass, toplevel,
				    XtNdefaultDistance, 2,
				    NULL);

        f2 = XtVaCreateManagedWidget("mmform0", formWidgetClass, f, 
				    XtNdefaultDistance, 1,
				     NULL);

        l = XtVaCreateManagedWidget("mmAddressl", labelWidgetClass, f2,
				    XtNresizable, TRUE,
				    XtNfromHoriz, NULL, 
				    NULL);

	lbuf0 = XtCalloc(MAX_BUFSIZE * 2, sizeof(char));
        a = XtVaCreateManagedWidget("mmInputLine", asciiTextWidgetClass, f2,
				    XtNfromHoriz, l,
				    XtNfromVert, NULL,
				    XtNresizable, TRUE,
				    XtNeditType, XawtextEdit, 
				    XtNtype, XawAsciiString,
				    XtNuseStringInPlace, True, 
				    XtNlength, MAX_BUFSIZE * 2, 
				    XtNstring, lbuf0, 
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNscrollVertical, XawtextScrollNever,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapWord,
				    NULL);
	typing_w = a;
	gbuf0 = lbuf0;
	XtAddCallback(a, XtNdestroyCallback, ReleaseString, lbuf0);
    	XtSetKeyboardFocus(f, a);

    	dial_w = XtVaCreateManagedWidget("mmDial", commandWidgetClass, f2,
					  XtNfromHoriz, a,
					  XtNfromVert, NULL,
					  XtNresizable, False,
					  XtNsensitive, True,
					  NULL);
	XtAddCallback(dial_w, XtNcallback, DialCallee, a);

    	option_w = XtVaCreateManagedWidget("mmOptions", commandWidgetClass, f2,
					  XtNfromHoriz, dial_w,
					  XtNfromVert, NULL,
					  XtNresizable, False,
					  XtNsensitive, True,
					  NULL);
	XtAddCallback(option_w, XtNcallback, PopupOptions, NULL);

    	logcommand_w = XtVaCreateManagedWidget("mmLogger", commandWidgetClass, f2,
					  XtNfromHoriz, option_w,
					  XtNfromVert, NULL,
					  XtNresizable, False,
					  XtNsensitive, True,
					  NULL);
	XtAddCallback(logcommand_w, XtNcallback, PopupHLog, NULL);

/*
	trans = XtParseTranslationTable(makecallTranslations);
	XtOverrideTranslations(a, trans);
*/

        f3 = XtVaCreateManagedWidget("mmform1", formWidgetClass, f,
				     XtNfromVert, f2,
				     XtNfromHoriz, NULL, 
				     XtNresizable, TRUE,
				     XtNdefaultDistance, 1,
				     NULL);

    	a = XtVaCreateManagedWidget("mmQuit", commandWidgetClass, f3,
			 	     XtNfromVert, NULL,
				     XtNfromHoriz, NULL,
					  NULL);
	XtAddCallback(a, XtNcallback, Quit, NULL);

    	accept_w = XtVaCreateManagedWidget("mmAccept", commandWidgetClass, f3,
					  XtNfromHoriz, a,
					  XtNfromVert, NULL,
					  XtNresizable, True,
					  XtNsensitive, False,
					  NULL);
	XtAddCallback(accept_w, XtNcallback, issueResponse, ACCEPT_CALL);

    	refuse_w = XtVaCreateManagedWidget("mmRefuse", commandWidgetClass, f3,
					  XtNfromHoriz, accept_w,
					  XtNfromVert, NULL,
					  XtNresizable, True,
					  XtNsensitive, False,
					  XtNradioGroup, accept_w,
					  NULL);
	XtAddCallback(refuse_w, XtNcallback, issueResponse, REFUSE_CALL);

    	hangup_w = XtVaCreateManagedWidget("mmHangup", commandWidgetClass, f3,
					  XtNfromHoriz, refuse_w,
					  XtNfromVert, NULL,
					  XtNresizable, True,
					  XtNsensitive, False,
					  XtNradioGroup, accept_w,
					  NULL);
	XtAddCallback(hangup_w, XtNcallback, hangUp, NULL);

    	putonhold_w = XtVaCreateManagedWidget("mmPutonhold", commandWidgetClass, f3,
					  XtNfromHoriz, hangup_w,
					  XtNresizable, True,
					  XtNfromVert, NULL,
					  XtNsensitive, False,
					  NULL);
	XtAddCallback(putonhold_w, XtNcallback, issueResponse, PUT_ONHOLD);

    	takemsg_w = XtVaCreateManagedWidget("mmTakemsg", commandWidgetClass, f3,
					  XtNfromHoriz, putonhold_w,
					  XtNresizable, True,
					  XtNfromVert, NULL,
					  XtNsensitive, False,
					  XtNradioGroup, accept_w,
					  NULL);
	XtAddCallback(takemsg_w, XtNcallback, issueResponse, TAKE_MSG);

    	pickcall_w = XtVaCreateManagedWidget("mmPickcall", menuButtonWidgetClass, f3,
					  XtNfromHoriz, takemsg_w,
					  XtNresizable, True,
					  XtNfromVert, NULL,
					  XtNsensitive, False,
					  XtNmenuName, "mmcallmenu",
					  NULL);

	callmenu_w = XtVaCreatePopupShell("mmcallmenu", simpleMenuWidgetClass, pickcall_w,
					  XtNlabel, "Calls On Hold",
					  NULL);
	
	isthereNewMsg = checkForMessages(0);
    	retrievemsg_w = XtVaCreateManagedWidget("mmRetrievemsg", commandWidgetClass, f3,
					  XtNfromHoriz, pickcall_w,
					  XtNresizable, True,
					  XtNfromVert, NULL,
					  XtNsensitive, True,
					  XtNlabel, (isthereNewMsg > 0 ? "  Read  Msg  ":
								    " Record Msg"),
					  NULL);
	XtAddCallback(retrievemsg_w, XtNcallback, PopupMessages, NULL);

        f4 = XtVaCreateManagedWidget("mmform2", formWidgetClass, f,
				     XtNfromVert, f3,
				     XtNfromHoriz, NULL, 
				     XtNdefaultDistance, 1,
				     NULL);

        l = XtVaCreateManagedWidget("mmStatus", labelWidgetClass, f4,
				    XtNresizable, TRUE,
				    XtNfromHoriz, NULL, 
				    NULL);

        status_w = XtVaCreateManagedWidget("mmStatusLine", asciiTextWidgetClass, f4,
				    XtNresizable, TRUE,
				    XtNfromHoriz, l,
				    XtNfromVert, NULL,
				    XtNdisplayCaret, FALSE, 
				    XtNeditType, XawtextEdit, 
				    XtNtype, XawAsciiString,
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNscrollVertical, XawtextScrollNever,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapNever,
				    NULL);

/*

	trans = XtParseTranslationTable(defaultTranslations);
	XtOverrideTranslations(status_w, trans);
*/

    	p = XtVaCreatePopupShell("Call History",  transientShellWidgetClass, toplevel, NULL);
        f = XtVaCreateManagedWidget("hForm0", formWidgetClass, p,
				    XtNdefaultDistance, 2,
				    XtNresizable, True,
				    XtNsensitive, True,
				    NULL);
        f2 = XtVaCreateManagedWidget("hForm1", formWidgetClass, f,
				    XtNdefaultDistance, 2,
				    XtNresizable, True,
				    XtNsensitive, True,
				    NULL);

    	a = XtVaCreateManagedWidget("hQuit", commandWidgetClass, f2,
				     XtNfromHoriz, NULL,
				     XtNresizable, True,
				     XtNsensitive, True,
				     NULL);
	XtAddCallback(a, XtNcallback, PopdownHLog, p);
	logQuit_w = a;

        l = XtVaCreateManagedWidget("hLabel", labelWidgetClass, f2,
				    XtNresizable, TRUE,
				    XtNfromVert, NULL,
				    XtNfromHoriz, a, 
				    XtNsensitive, True,
				    NULL);
	logLabel_w = l;

        b = XtVaCreateManagedWidget("hList", asciiTextWidgetClass, f,
				    XtNresizable, TRUE,
				    XtNfromHoriz, NULL,
				    XtNfromVert, f2,
				    XtNleftMargin, 4,
				    XtNtopMargin, 4,
				    XtNrightMargin, 4,
				    XtNbottomMargin, 2,
				    XtNdisplayCaret, True, 
				    XtNeditType, XawtextAppend, 
				    XtNtype, XawAsciiString,
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNscrollVertical, XawtextScrollAlways,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapNever,
				    NULL);
	hisList_w = b;
	hisLog_w = p;
    	return (f);
}	



Widget  create_talkpanel(shellname)
	char	*shellname;
{
	char i;
	Widget p;
	struct tm *t;
	sInputRec *ir;
	struct timeval ct;
	XawTextBlock text;
	char buf[MAX_MSGBUF];
    	Widget toplevel, f, f2, f3, f4, a, b, l;
    
	if (panel_counter == Max_Panels)
	    return (NULL);

	if ((ir = (sInputRec *)XtMalloc(sizeof(sInputRec))) == NULL) {
	    issueError("  Malloc failed -- too many talk panels, maybe.", SHORT_REPEAT, FQ7_INDX, 
			BELL_TONE0, 2);
	    return (NULL);
	}

	for (i = 0; i < Max_Panels; i++)
	     if (talkpanels[i].id == -1) break;

	if (get_tsocket(ir) < 0) {
	    XtFree((char *)ir);
	    return (NULL);
	}

    	toplevel = XtVaAppCreateShell(shellname, (String)applicationShellWidgetClass,
				      (String)topLevelShellWidgetClass, mydpy,
				      XtNiconPixmap, TalklinePixmap,
				      XtNiconName, shellname,
				      NULL);

    	p = XtVaCreatePopupShell("File",  overrideShellWidgetClass, toplevel, NULL);
        f = XtVaCreateManagedWidget("RForm", formWidgetClass, p,
				    XtNdefaultDistance, 2,
				    XtNresizable, False,
				    NULL);

        l = XtVaCreateManagedWidget("fLabel", labelWidgetClass, f,
				    XtNresizable, False,
				    XtNfromHoriz, NULL, 
				    NULL);

        a = XtVaCreateManagedWidget("fWindow", asciiTextWidgetClass, f,
				    XtNfromHoriz, l,
				    XtNfromVert, NULL,
				    XtNresizable, False,
				    XtNdisplayCaret, True, 
				    XtNtype, XawAsciiString,
				    XtNeditType, XawtextEdit, 
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNscrollVertical, XawtextScrollNever,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapWord,
				    NULL);
	ir->fname_w = a;
    	b = XtVaCreateManagedWidget("fOkay", commandWidgetClass, f,
				    XtNfromHoriz, a,
				    XtNfromVert, NULL,
				    XtNresizable, False,
				    NULL);
	XtAddCallback(b, XtNcallback, OkayOpenFile, a);

    	a = XtVaCreateManagedWidget("fCancel", commandWidgetClass, f,
				    XtNfromHoriz, b,
				    XtNfromVert, NULL,
				    XtNresizable, False,
				    NULL);
	XtAddCallback(a, XtNcallback, CancelOpenFile, ir);
	ir->fd = -1;
	ir->fw = p;
	ir->recording = 0;
	ir->buf = ir->epos = ir->bp = (char *)0;

		            
        f = XtVaCreateManagedWidget("talkpanel", formWidgetClass, toplevel,
				    XtNdefaultDistance, 2,
				    XtNresizable, True,
				    NULL);
	XtAddCallback(f, XtNdestroyCallback, ReleaseInputRec, ir);

        f2 = XtVaCreateManagedWidget("tCommandPanel", formWidgetClass, f,
				     XtNfromVert, NULL,
				     XtNfromHoriz, NULL, 
				     XtNresizable, TRUE,
				     XtNdefaultDistance, 1,
				     NULL);

    	b = XtVaCreateManagedWidget("tQuit", commandWidgetClass, f2,
					  XtNfromHoriz, NULL,
					  XtNfromVert, NULL,
					  XtNresizable, TRUE,
					  NULL);
	XtAddCallback(b, XtNcallback, Popdowntalkpanel, i);

    	a = XtVaCreateManagedWidget("tFile", commandWidgetClass, f2,
					  XtNfromHoriz, b,
					  XtNfromVert, NULL,
					  XtNresizable, TRUE,
					  NULL);
	XtAddCallback(a, XtNcallback, PopupFilepanel, ir);
	ir->fb = a;

    	b = XtVaCreateManagedWidget("tSRecord", toggleWidgetClass, f2,
					  XtNfromHoriz, a,
					  XtNfromVert, NULL,
/*
					  XtNresizable, False,
*/
					  XtNsensitive, False,
					  XtNlabel, " Record ",
					  XtNstate, True,
					  NULL);
	XtAddCallback(b, XtNcallback, StopRecording, ir);
	ir->rb = b;

	(void)gettimeofday(&ct, (struct timezone *)0);
	t = localtime(&ct.tv_sec);
	(void)sprintf(buf, " %-s@%-s    %02d:%02d:%02d  %2d/%02d/%04d", hisNameBuf, hisAddressBuf, 
		t->tm_hour, t->tm_min, t->tm_sec, t->tm_mon + 1, t->tm_mday, t->tm_year + 1900);
	

        a = XtVaCreateManagedWidget("tHisinfo", asciiTextWidgetClass, f2,
				    XtNfromHoriz, b,
				    XtNfromVert, NULL,
				    XtNresizable, TRUE,
				    XtNdisplayCaret, FALSE, 
				    XtNtype, XawAsciiString,
				    XtNeditType, XawtextEdit, 
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNscrollVertical, XawtextScrollNever,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapWord,
				    NULL);

	text.length = strlen(buf);
	text.firstPos = 0;
	text.ptr = buf;
	(void)XawTextReplace(a, 0, 0, &text);

/*
	trans = XtParseTranslationTable(hiswinTranslations);
	XtOverrideTranslations(a, trans);
*/

        f3 = XtVaCreateManagedWidget("tHispanelf", formWidgetClass, f,
				     XtNfromVert, f2,
				     XtNfromHoriz, NULL, 
				     XtNresizable, TRUE,
				     XtNdefaultDistance, 1,
				     NULL);

        a = XtVaCreateManagedWidget("tHtalkpanel", asciiTextWidgetClass, f3,
				    XtNresizable, TRUE,
				    XtNfromVert, NULL,
				    XtNfromHoriz, NULL,
				    XtNeditType, XawtextEdit, 
				    XtNtype, XawAsciiString,
				    XtNdisplayCaret, True, 
				    XtNscrollVertical, XawtextScrollAlways,
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNresize, XawtextResizeNever,
				    XtNwrap, XawtextWrapWord,
				    NULL);
	ir->w0 = a;
/*
	trans = XtParseTranslationTable(hiswinTranslations);
	XtOverrideTranslations(a, trans);
*/

        b = XtVaCreateManagedWidget("tMytalkpanel", asciiTextWidgetClass, f3,
				    XtNfromVert, a,
				    XtNfromHoriz, NULL,
				    XtNresizable, TRUE,
				    XtNeditType, XawtextEdit, 
				    XtNtype, XawAsciiString,
				    XtNdisplayCaret, True, 
				    XtNscrollVertical, XawtextScrollAlways,
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNwrap, XawtextWrapWord,
				    NULL);
	ir->w1 = b;
/*
	trans = XtParseTranslationTable(defaultTranslations1);
	XtOverrideTranslations(b, trans);
*/

        f4 = XtVaCreateManagedWidget("tMypanelf", formWidgetClass, f,
				     XtNfromVert, f3,
				     XtNfromHoriz, NULL, 
				     XtNresizable, TRUE,
				     XtNdefaultDistance, 1,
				     NULL);

        a = XtVaCreateManagedWidget("tMytypingpanel", asciiTextWidgetClass, f4,
				    XtNresizable, TRUE,
				    XtNfromVert, NULL,
				    XtNfromHoriz, NULL,
				    XtNtype, XawAsciiString,
				    XtNeditType, XawtextEdit, 
				    XtNscrollVertical, XawtextScrollNever,
				    XtNscrollHorizontal, XawtextScrollNever,
				    XtNwrap, XawtextWrapNever,
				    NULL);
    	XtSetKeyboardFocus(f, a);
	ir->w2 = a;

	ir->inputId = XtAppAddInput(xcomm, ir->sockt, XtInputReadMask, ReadCharFromHim, ir);
	ir->exceptionId = XtAppAddInput(xcomm, ir->sockt, XtInputExceptMask, CloseTalkSocket, ir);

/*
	trans = XtParseTranslationTable(mywinTranslations);
	XtOverrideTranslations(a, trans);
*/
	ir->pw = toplevel;
	talkpanels[i].w = toplevel;
	talkpanels[i].info = ir;
	talkpanels[i].id = panel_counter++;

    	return (toplevel);
}	


/*ARGSUSED*/
static
void Quit(w, event, params, num_params)
	Widget w;
	XButtonEvent *event;
	String *params;
	Cardinal *num_params;
{
	register int i;

	XtRemoveInput(relayId);
	XtRemoveInput(rexceptionId);
	for (i = 0; i < MAX_QUE_CALLS; i++)
	    if (talkpanels[i].id != -1) {
		close(talkpanels[i].info->sockt);
		if (talkpanels[i].info->fd > 0) close(talkpanels[i].info->fd);
	    }
	(void)kill(childpid, SIGUSR2);
}

/*
	create file to write to 
*/
/*ARGSUSED*/
static
void OpenRFile(w, event, params, num_params)
	Widget w;
	XButtonEvent *event;
	String *params;
	Cardinal *num_params;
{
	OpenRecordingFile(w);
}

/*ARGSUSED*/
static
void	OkayOpenFile(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	Widget tw = (Widget)client_data;

	OpenRecordingFile(tw);
}

void OpenRecordingFile(w)
	Widget w;
{
	Widget s;
	int i, fd;
	register char *sp;
	XawTextBlock text;
	XawTextPosition ipos;


	if ((ipos = XawTextGetInsertionPoint(w)) > 0) {
	    text.length = 0;
	    text.firstPos = 0;
	    text.ptr = (char *)0;
	    s = XawTextGetSource(w);
	    XawTextSourceRead(s, 0, &text, ipos);
	    for (sp = text.ptr; isspace(*sp); sp++);
	}

	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->fname_w == w) {
		    if (ipos <= 0) break;
	 	    if ((fd = open(sp, (O_CREAT | O_RDWR | O_APPEND), 0666)) >= 0) {
			sInputRec *ir = talkpanels[i].info;

			if (ir->recording) {
			    if (ir->bp != ir->buf) {
			        *(ir->bp) = '\0';
			        (void)write(ir->fd, ir->buf, strlen(ir->buf));
			        ir->bp = ir->buf;
			    }
			    (void)close(ir->fd);
			}

			ir->fd = fd;
			XtSetSensitive(ir->rb, True);
			ir->recording = 1;
			if (ir->buf == NULL) {
			    ir->buf = ir->bp = XtCalloc(MAX_BUFSIZE * 2, sizeof(char));
			    ir->epos = &ir->buf[(MAX_BUFSIZE * 2) - 2];
			}
		    }
		    else {
			char buf[MAX_MSGBUF];

			(void)sprintf(buf, "  Error opening >%-s<", sp);
			issueError(buf, VVSHORT_REPEAT, FQ5_INDX, BELL_TONE3, 1);
		    }
		    break;
	        }
        XtPopdown(talkpanels[i].info->fw);
}

/*ARGSUSED*/
static
void	StopRecording(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	sInputRec *ir = (sInputRec *)client_data;

	XtVaSetValues(ir->rb, XtNresizable, False,
		      XtNlabel, ir->recording ? " Pause  ":" Record ",
		      NULL);
	ir->recording = ir->recording ? 0:1;
	XtVaSetValues(ir->rb, XtNresizable, True, NULL);
}

/*ARGSUSED*/
static
void	PopupFilepanel(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	Position x, y;
	Dimension width, height;
	sInputRec *ir = (sInputRec *)client_data;

	XtVaGetValues(ir->pw,
		XtNx, &x,
		XtNy, &y,
		XtNwidth, &width,
		XtNheight, &height,
		NULL);
/*
	XtTranslateCoords(ir->fw, (Position) width/2, (Position) height/2,
			  &x, &y);
*/
	XtVaSetValues(ir->fw, 
		XtNx, x + 35,
		XtNy, y + (height/2) + 17,
		NULL);
	XtPopup(ir->fw, XtGrabNonexclusive);
}

/*ARGSUSED*/
static
void	CancelOpenFile(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	sInputRec *ir = (sInputRec *)client_data;

        XtPopdown(ir->fw);

}

/*ARGSUSED*/
static
void	ReleaseInputRec(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	sInputRec *ir = (sInputRec *)client_data;
/*
	XtSetSensitive(ir->w1, False);
	XtSetSensitive(ir->fb, False);
	XtSetSensitive(ir->rb, False);
*/
	XtRemoveInput(ir->inputId);
	XtRemoveInput(ir->exceptionId);
	(void)write(ir->sockt, &quitCode, 1);
	(void)close(ir->sockt);
	if (ir->buf != ir->bp) {
	    *(ir->bp) = '\0';
	    (void)write(ir->fd, ir->buf, strlen(ir->buf));
	    (void)close(ir->fd);
	}
	if (ir->buf != (char *)0)
	    XtFree((char *)ir->buf);
	XtFree((char *)ir);
}


/*ARGSUSED*/
static
void	ReleaseString(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	String sp = (String)client_data;

	XtFree((char *)sp);
}


static
void	getTalkLine()
{
	Widget p;
	Bool error = True;
	static char buf[MAX_MSGBUF];

	msgAnnouncer = False;
	if (panel_counter < Max_Panels) {
	    (void)sprintf(buf, "%-s@%-s", hisNameBuf, hisAddressBuf);
	    msgAnnouncer = False;
	    error = False;
	    if ((p = create_talkpanel(buf)) != NULL) {
	       XtRealizeWidget(p);
	       XtSetSensitive(typing_w, True);
	       XtSetSensitive(dial_w, True);
	       if (callwaiting > 0)
		   XtSetSensitive(pickcall_w, True);
	    }
	}
	if (error) {
	    apoligize();
	    issueError("  New talk session not made, you have too many talk sessions.", VSHORT_REPEAT, FQ5_INDX, BELL_TONE2, 2);
	}
}

void	apoligize()
{
	sInputRec ir;

	if (get_tsocket(&ir) < 0)
	    return;
	(void)write(ir.sockt, sorryMsg, strlen(sorryMsg));
	(void)write(ir.sockt, QUIT_CODE, 1);
	(void)close(ir.sockt);
}

/*ARGSUSED*/
static
void	CloseRelayLine(client_data, source, id)
	XtPointer client_data;
	int *source;
	XtInputId *id;
{
	int status;
	(void)fprintf(stderr, "Relay connection (socket) just failed for some reason.  Heading for the ...");

	ring_bell(30);
	XtRemoveInput(relayId);
	XtRemoveInput(rexceptionId);
	XtRemoveInput(talkLineId);
	XtRemoveInput(texceptionId);
	(void)kill(childpid, SIGUSR2);
}

/*ARGSUSED*/
static
void	CloseTalkLine(client_data, source, id)
	XtPointer client_data;
	int *source;
	XtInputId *id;
{
	int status;
	(void)fprintf(stderr, "Talk connection (socket) just failed for some reason.  Heading out ...");

	ring_bell(30);
	XtRemoveInput(relayId);
	XtRemoveInput(rexceptionId);
	XtRemoveInput(talkLineId);
	XtRemoveInput(texceptionId);
	(void)kill(childpid, SIGUSR2);
}


/*ARGSUSED*/
static
void	Popdowntalkpanel(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	int id = (int)client_data;

	panel_counter--;
	talkpanels[id].id = -1;
/*
        XtPopdown(talkpanels[id].w);
*/
	XtUnrealizeWidget(talkpanels[id].w);
	XtDestroyWidget(talkpanels[id].info->fw);
	XtDestroyWidget(talkpanels[id].w);

}

/*ARGSUSED*/
static
void Beep(w, event, params, num_params)
	Widget w;
	XButtonEvent *event;
	String *params;
	Cardinal *num_params;
{
    	ring_bell(1);
}

void	ring_bell(duration)
	int  duration;
{
    	XBell(mydpy, duration);
}

/*ARGSUSED*/
static
void	issueResponse(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	int resp = (int)client_data;


	XtRemoveTimeOut(requestId);
	XtRemoveTimeOut(callIndicatorID);
	XtVaSetValues(topLevel, XtNiconPixmap, icon_pixmap, NULL);
	(void)xrelay_msg(&pc_msg, resp);
	msgAnnouncer = False;
	XtSetSensitive(typing_w, False);
	XtSetSensitive(dial_w, False);
	XtSetSensitive(accept_w, False);
	XtSetSensitive(refuse_w, False);
	XtSetSensitive(putonhold_w, False);
	XtSetSensitive(takemsg_w, False);
	if (resp != ACCEPT_CALL) {
	    if (callwaiting > 0)
		XtSetSensitive(pickcall_w, True);
/*
	    XtOverrideTranslations(typing_w, typingTrans);
*/
	    XtSetSensitive(typing_w, True);
	    XtSetSensitive(dial_w, True);
	    clear_textwin(typing_w);
	}

}

/*ARGSUSED*/
static
void	DialCallee(w, client_data, call_data)
	Widget w;
	XtPointer client_data;
	XtPointer call_data;
{
	Widget tw = (Widget)client_data;

	parseCMDLine(tw);
}

/*ARGSUSED*/
static
void MakeCall(w, event, params, num_params)	/* get string from typing window in */
	Widget	w;				/* main menu and pass it to the child */
	XButtonEvent *event;			/* to do the calling. The child will pass */
	String *params;				/* back information about the callee before */
	Cardinal *num_params;			/* it actually begins the calling process */
{						
	parseCMDLine(typing_w);
}

void parseCMDLine(w)
	Widget w;
{
	LOCAL_MSG lmsg;
	Bool error = True;
	register String sp;
	XawTextPosition ipos;

	msgAnnouncer = False;
	ipos = XawTextGetInsertionPoint(w);

	if (ipos > 0) {
	    sp = gbuf0;
	    sp[ipos] = '\0';
	    for (; isspace(*sp); sp++, ipos--);
	    if (ipos > 0) {
		error = False;
		(void)sprintf(lmsg.line, "%-s", sp);
	    }
	}

	if (!error && !xrelay_msg(&lmsg, MAKE_CALL)) {
	    XtSetSensitive(hangup_w, True);
	    XtSetSensitive(typing_w, False);
	    XtSetSensitive(dial_w, False);
/*
	    XtOverrideTranslations(typing_w, notypingTrans);
*/
	    if (callwaiting > 0)
		XtSetSensitive(pickcall_w, False);
	}
	else {
	    ring_bell(10);
	    clear_textwin(typing_w);
	}

	return;

}


void	clear_textwin(w)
	Widget w;
{
	XawTextPosition ipos;
	XawTextBlock text;

	text.length = 0;
	text.firstPos = 0;
	text.ptr = (char *)0;

	ipos = XawTextGetInsertionPoint(w);
	(void)XawTextReplace(w, 0, ipos, &text);
}

/*ARGSUSED*/
static
void SendKillLine(w, event, params, nparams)
    	Widget w;
    	XEvent *event;
    	String *params;
    	Cardinal *nparams;
{
	int nb;
	register int i;

	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->w2 == w) {
		    if (XawTextGetInsertionPoint(w) > 0) {
	                nb = write(talkpanels[i].info->sockt, my_edchars + 1, 1);
	                if (nb != 1) 
	                    issueError("  Error trying to relay line kill.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);
		    }
		    return;
	        }
}

/*ARGSUSED*/
static
void SendDeleteChar(w, event, params, nparams)
    	Widget w;
    	XEvent *event;
    	String *params;
    	Cardinal *nparams;
{
	int nb;
	register int i;

	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->w2 == w) {
		    if (XawTextGetInsertionPoint(w) > 0) {
	                nb = write(talkpanels[i].info->sockt, my_edchars, 1);
	                if (nb != 1) 
	                    issueError("  Error trying to relay character delete.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);
		    }
		    return;
		}
}


/*ARGSUSED*/
static
void SendDeleteWord(w, event, params, nparams)
    	Widget w;
    	XEvent *event;
    	String *params;
    	Cardinal *nparams;
{
	int nb;
	register int i;

	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->w2 == w) {
		    if (XawTextGetInsertionPoint(w) > 0) {
	                nb = write(talkpanels[i].info->sockt, my_edchars + 2, 1);
	                if (nb != 1) 
	                    issueError("  Error trying to relay word delete.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);
		    }
		    return;
		}

}


/*ARGSUSED*/
static
void SendCharToHim(w, event, params, nparams)
    	Widget w;
    	XEvent *event;
    	String *params;
    	Cardinal *nparams;
{
	int len, nb;
	register int i;
	XButtonEvent *bevent;
	KeySym keysym_return;
	XComposeStatus status_in_out;
	static char buf[Little_BufSize];

	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->w2 == w) {
		    buf[0] = '\0';
		    bevent = &event->xbutton;
	            len = XLookupString((XKeyEvent *)bevent, buf, Little_BufSize, &keysym_return, &status_in_out);
	            if (!len || keysym_return == XK_Shift_L || keysym_return == XK_Shift_R) return;

		    nb = write(talkpanels[i].info->sockt, buf, len); 
	            if (nb != len) 
	                issueError("  Error trying to relay character.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);

		    return;
		}
	/* shouldn't reach this point */
	issueError("  Xcomm out of sync --- might want to restart !!!", SHORT_REPEAT, FQ5_INDX, BELL_TONE4, INFINITE_RING);
}

/*ARGSUSED*/
void	getSelection(w, client_data, selection, type, value, length, format)
	Widget w;
	XtPointer client_data;
	Atom *selection;
	Atom *type;
	XtPointer value;
	unsigned long *length;
	int *format;
{
	int ch;
	Widget s;
	register int i;
	XawTextBlock text;
	register char *fsp;
	XawTextPosition ipos, ipos2;

	String sp = (String)value;

	if (sp != NULL && *length != 0) {
	    for (i = 0; i < Max_Panels; i++)
	         if (talkpanels[i].id != -1)
		     if (talkpanels[i].info->w2 == w) {
			 text.firstPos = 0;
			 text.ptr = (char *)0;
			 text.length = 0;
			 if ((fsp = strrchr(sp, '\n')) != (char *)0) {
			     ch = *++fsp; *fsp = '\0';
	                     ipos = XawTextGetInsertionPoint(talkpanels[i].info->w2);

			     if (ipos > 0) {
			         s = XawTextGetSource(talkpanels[i].info->w2);
			         if (XawTextSourceRead(s, 0, &text, ipos) > 0) {
	                             ipos2 = XawTextGetInsertionPoint(talkpanels[i].info->w1);
			             XawTextReplace(talkpanels[i].info->w1, ipos2, ipos2, &text);
	                             XawTextSetInsertionPoint(talkpanels[i].info->w1, text.length + ipos2);
			 	     text.length = 0;
			 	     text.firstPos = 0;
			 	     text.ptr = (char *)0;
			             XawTextReplace(talkpanels[i].info->w2, 0, ipos, &text);
			         }
			     }

			     text.ptr = sp;
			     text.firstPos = 0;
			     text.length = strlen(sp);
	                     ipos2 = XawTextGetInsertionPoint(talkpanels[i].info->w1);
			     XawTextReplace(talkpanels[i].info->w1, ipos2, ipos2, &text);
	                     XawTextSetInsertionPoint(talkpanels[i].info->w1, text.length + ipos2);
			     *fsp = ch;
			 }
	                 text.firstPos = 0;
	                 text.ptr = (fsp == NULL) ? sp:fsp;
	                 text.length = (fsp == NULL) ? *length:strlen(fsp);
	                 ipos = XawTextGetInsertionPoint(talkpanels[i].info->w2);
	                 (void)XawTextReplace(talkpanels[i].info->w2, ipos, ipos, &text);
	                 XawTextSetInsertionPoint(talkpanels[i].info->w2, text.length + ipos);

		         if (write(talkpanels[i].info->sockt, sp, *length) != *length)
	                     issueError("  Error relaying selection.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);

		     }
	}
	XtFree(value);
}


/*ARGSUSED*/
void 	RequestSelection(w, tevent, params, nparams)
	Widget w;
	XEvent *tevent;
	String *params;
	Cardinal *nparams;
{
	XButtonEvent *event = (XButtonEvent *)tevent;

	XtGetSelectionValue(w, XA_PRIMARY, XA_STRING, getSelection, event, event->time);
}


/*ARGSUSED*/
void 	TransferLine(w,event, params, nparams)
	Widget w;
	XEvent *event;
	String *params;
	Cardinal *nparams;
{
	Widget s;
	register int i;
	XawTextBlock text;
	KeySym keysym_return;
	int cread;
	XComposeStatus status_in_out;
	static char buf[MAX_MSGBUF];
	XawTextPosition rpos, ipos;


	for (i = 0; i < Max_Panels; i++)
	    if (talkpanels[i].id != -1)
		if (talkpanels[i].info->w2 == w) {
	            text.firstPos = 0;
		    text.ptr = buf;
		    text.length = MAX_BUFSIZE;

		    ipos = XawTextGetInsertionPoint(w);
		    s = XawTextGetSource(w);

		    cread = XawTextSourceRead(s, 0, &text, MAX_MSGBUF);

		    text.ptr[cread] = '\n';
		    text.ptr[cread + 1] = '\0';
		    text.length++;
		    rpos = XawTextGetInsertionPoint(talkpanels[i].info->w1);
		    (void)XawTextReplace(talkpanels[i].info->w1, rpos, rpos, &text);
		    XawTextSetInsertionPoint(talkpanels[i].info->w1, rpos + cread + 1);

		    text.ptr = (char *)0;
		    text.length = 0;
		    (void)XawTextReplace(w, 0, ipos, &text);
		    XawTextSetInsertionPoint(w, 0);


		    if (write(talkpanels[i].info->sockt, "\n", 1) != 1)
	                issueError("  Error while trying to relay newline.", VVSHORT_REPEAT, FQ5_INDX, BELL_TONE4, 1);

		    return;
	  	}
	/* shouldn't reach this point */
	issueError("  Xcomm out of sync --- might want to restart !!!", SHORT_REPEAT, FQ6_INDX, BELL_TONE4, INFINITE_RING);
}

int	xrelay_msg(msg, msg_type)
	LOCAL_MSG *msg;
	int  msg_type;
{
	int  nb;
	static int errorc = 0;


	msg->type = msg_type;
	nb = write(fds[0], (char *)msg, sizeof(*msg));
	if (nb != sizeof(*msg)) {
	    issueError("  Relay connection (socket) is experiencing difficulties.", SHORT_REPEAT,
		       FQ5_INDX, BELL_TONE4, 3);
	    if (++errorc > MAX_RELAYERRORS) {
		(void)fprintf(stderr, "Errors on relay socket: %3d times\n", errorc);
	        issueError("  Errors on relay socket !!!", SHORT_REPEAT, FQ5_INDX, BELL_TONE5, INFINITE_RING);
		(void)kill(childpid, SIGUSR2);
	    }
	    return (-1);
	}
	return (0);
}


/*ARGSUSED*/
static
void	CloseTalkSocket(client_data, source, id)
	XtPointer client_data;
	int *source;
	XtInputId *id;
{
	sInputRec  *ir = (sInputRec *)client_data;
/*
	XtSetSensitive(ir->w1, False);
	XtSetSensitive(ir->rb, False);
	XtSetSensitive(ir->fb, False);
*/
	XtRemoveInput(ir->inputId);
	XtRemoveInput(ir->exceptionId);
	(void)write(ir->sockt, &quitCode, 1);
	(void)close(ir->sockt);
}

/*ARGSUSED*/
static
void	ReadCharFromHim(client_data, source, id)
	XtPointer client_data;
	int *source;
	XtInputId *id;
{
	int cc;
	Bool updatePos;
	register int i;
	XawTextBlock text;
 	char buf[Little_BufSize];
	register char *sp = buf;
	XawTextPosition spos, epos;
	sInputRec  *ir = (sInputRec *)client_data;

	if ((cc = read(*source, buf, Little_BufSize)) > 0) {
	    if (buf[0] == QUIT_CODE || buf[0] < 4) {
		if (ir->recording && ir->buf != ir->bp) {
		    *(ir->bp) = '\0';
		    (void)write(ir->fd, ir->buf, strlen(ir->buf));
		    ir->bp = (char *)0;
		}
	        XtVaSetValues(ir->pw, XtNiconPixmap, iTalklinePixmap, NULL);
		XtSetSensitive(ir->w2, False);
		XtSetSensitive(ir->rb, False);
		XtSetSensitive(ir->fb, False);
		return;
	    }
	    buf[cc] = '\0';
	    for (i = 0; i < cc; i++, sp++) {
		updatePos = FALSE;
		epos = spos = XawTextGetInsertionPoint(ir->w0);

		if (*sp == ir->edchars[0] || *sp == ir->edchars[1] || *sp == ir->edchars[2]) {
		    if (ir->recording)
	                EditHisinput(*sp, ir->edchars, ir->buf, &ir->bp);

		    if (*sp == ir->edchars[0])
			spos--;
		    else {
			Widget s;
			s = XawTextGetSource(ir->w0);
			spos = XawTextSourceScan(s, epos, (*sp == ir->edchars[2]) ? 
						 XawstWhiteSpace:XawstEOL, XawsdLeft, 1, False); 
		    }
		    text.ptr = (char *)0;
		    text.length = 0;
		}
		else {
		    text.ptr = sp;
		    text.length = 1;
		    updatePos = True;
		    if (ir->recording) {
		        *(ir->bp)++ = *sp;
			if (*sp == '\n' || ir->bp >= ir->epos) {
			    *(ir->bp) = '\0';
			    (void)write(ir->fd, ir->buf, strlen(ir->buf));
			    ir->bp = ir->buf;
			}
		    }
		}
		text.firstPos = 0;
		XawTextReplace(ir->w0, spos, epos, &text);
/*
		if (!updatePos)
		    XawTextSetInsertionPoint(ir->w0, spos);
		else
*/
		   XawTextSetInsertionPoint(ir->w0,  updatePos ? (XawTextGetInsertionPoint(ir->w0) + 1):spos);
	   }
	}
	else {
	    if (ir->recording && ir->buf != ir->bp) {
		*(ir->bp) = '\0';
		(void)write(ir->fd, ir->buf, strlen(ir->buf));
		ir->bp = (char *)0;
	    }
/*
	    XtSetSensitive(ir->w2, False);
	    XtSetSensitive(ir->fb, False);
	    XtSetSensitive(ir->rb, False);
*/
	    XtVaSetValues(ir->pw, XtNiconPixmap, iTalklinePixmap, NULL);
	    return;
	}
}


int	get_tsocket(info)
	sInputRec *info;
{
	fd_set readmask;
	struct timeval timeout;
	static struct msghdr msg;
	static struct iovec iov[1];

	iov[0].iov_base = (char *)0;
	iov[0].iov_len = 0;
	
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;				/* get the talk socket from child */
	msg.msg_name = (caddr_t)0;
	msg.msg_accrights = (caddr_t)&info->sockt;
	msg.msg_accrightslen = sizeof(info->sockt);

	XtSetSensitive(hangup_w, False);
	if (recvmsg(filed_con[0], &msg, 0) < 0) {
	    issueError("  Error getting talk socket !!!", VSHORT_REPEAT, FQ6_INDX, BELL_TONE2, 3);
	    return (-1);
	}

        (void)write(info->sockt, my_edchars, 3); 		/* send him my editing characters */

	errno = 0;

	FD_ZERO(&readmask);
	FD_SET(info->sockt, &readmask);

	timeout.tv_usec = 0;
	timeout.tv_sec = 30;
	if (select(info->sockt + 1, &readmask, (fd_set *)0, (fd_set *)0, &timeout) <= 0) {
	    char buf[MAX_MSGBUF];

	    (void)close(info->sockt);
	    (void)sprintf(buf, "  %-s xcomm didn't quite make it !!!", hisNameBuf);
	    issueError(buf, VSHORT_REPEAT, FQ5_INDX, BELL_TONE2, 2);
	    return (-1);
	}
        read(info->sockt, info->edchars, 3); 
	return (info->sockt);
}

/*ARGSUSED*/
void	changeIcon(client_data, id)
	XtPointer client_data;
	XtIntervalId *id;
{
	XtVaSetValues(topLevel, XtNiconPixmap, solidIcon ? icon_pixmap:iicon_pixmap, NULL);
	callIndicatorID = XtAppAddTimeOut(xcomm, msgIntervals[FQ4_INDX], changeIcon, NULL);
	solidIcon = solidIcon ? False:True;
}

void	issueError(estring, repeatc, intervalIdx, bduration, ringcount)
	char *estring;
	int repeatc;
	int intervalIdx;
	char bduration;
	int ringcount;
{
	XawTextBlock text;
	XawTextPosition epos;
	errorRec *ep = &gerrordata;

	if (ep->active) {
	    XtRemoveTimeOut(ep->id);
	    XtFree(ep->estring);
	}
	ep->len = strlen(estring);
	ep->idx = intervalIdx;
	ep->repeatc = repeatc + 1;
	if (ringcount)
	    ep->ringcount = ringcount - 1;
	ep->bellduration = bduration;
	ep->estring = XtNewString(estring);
	repeatCount = repeatc;
	msgAnnouncer = True;
	ep->active = True;
	text.firstPos = 0;
	text.length = ep->len;
	text.ptr = ep->estring;
	if (ringBell && ringcount) 
	    ring_bell(ep->bellduration);
	epos = XawTextGetInsertionPoint(status_w);
	if (epos < ep->len) epos = ep->len;
	XawTextReplace(status_w, 0, epos, &text);
	XawTextSetInsertionPoint(status_w, epos);
	ep->id = XtAppAddTimeOut(xcomm, msgIntervals[ep->idx], msgRepeater, ep);
}

/*ARGSUSED*/
static
void	msgRepeater(client_data, id)
	XtPointer client_data;
	XtIntervalId *id;
{

	XawTextBlock text;
	static char status = 1;
	XawTextPosition epos;
	errorRec *ep = (errorRec *)client_data;

	text.firstPos = 0;
	if (msgAnnouncer && !status && (ep->repeatc -1) >= 0) {
	    --ep->repeatc;
	    text.length = ep->len;
	    text.ptr = ep->estring;
	    if (ep->ringcount > 0 && ringBell) {
	        ring_bell(ep->bellduration);
		ep->ringcount--;
	    }
	}
	else {
	    if (!status && ep->active) ep->repeatc--;
	    text.length = 0;
	    text.ptr = (char *)0;
	}

	epos = XawTextGetInsertionPoint(status_w);
	if (epos < ep->len)
	    epos = ep->len;

	XawTextReplace(status_w, 0, epos, &text);
	XawTextSetInsertionPoint(status_w, epos);

	if (msgAnnouncer && ep->repeatc >= 0) {
	    status = (status) ? 0:1;
	    ep->id = XtAppAddTimeOut(xcomm, msgIntervals[ep->idx], msgRepeater, ep);
	}
	else {
	    status = 1;
	    if (ep->active) {
	        XtFree(ep->estring);
		ep->active = False;
	    }
	    clear_textwin(status_w);
	}
}
	    

void	issueRequest(msg, buf, trequest)
	LOCAL_MSG *msg;
	char buf[];
	Bool trequest;
{
	register char *sp;
	char *line;

	msgAnnouncer = False;
	line = msg->line;
	hisAddressBuf[0] = hisTtyBuf[0] = '\0';
	sp = strtok(line, "~");
	if (*sp) (void)strcpy(hisNameBuf, sp);
	sp = strtok((char *)0, "~");
	if (*sp) (void)strcpy(hisAddressBuf, sp);
	sp = strtok((char *)0, "~");
	if (sp && *sp) (void)strcpy(hisTtyBuf, sp);
	if (trequest) {
	    struct tm *t;
	    struct timeval ct;
	    XawTextBlock text;
	    XawTextPosition ipos;
	    extern struct tm *localtime();
	
	    gettimeofday(&ct, (struct timezone *)0);
	    t = localtime(&ct.tv_sec);

	    sprintf(buf, "%02d:%02d  %02d/%02d  %-s@%-s\n", t->tm_hour, t->tm_min, t->tm_mon + 1, t->tm_mday,
		    hisNameBuf, hisAddressBuf);

	    text.firstPos = 0;
	    text.ptr = buf;
	    text.length = strlen(buf);

	    if (!logCounter) logCounter = 1;
	    ipos = XawTextGetInsertionPoint(hisList_w);
	    XawTextReplace(hisList_w, ipos, ipos, &text);
	    XawTextSetInsertionPoint(hisList_w, ipos + text.length);

	    (void)sprintf(buf, "  talk request from %-s%-s%-s%-s%-s", hisNameBuf, hisAddressBuf[0] ? "@":
			 " ", hisAddressBuf[0] ? hisAddressBuf:"", hisTtyBuf[0] ? "on tty ":"", 
			 hisTtyBuf[0] ? hisTtyBuf:"");

	    requestId = XtAppAddTimeOut(xcomm, INVITATION_INTVAL, invitationRepeater, NULL);

	    XtSetSensitive(typing_w, False);
	    XtSetSensitive(dial_w, False);
/*
	    XtSetSensitive(hangup_w, True);
*/
	    XtSetSensitive(accept_w, True);
	    XtSetSensitive(refuse_w, True);
	    if (msg->id == XCOMMFLAG)
	        XtSetSensitive(putonhold_w, True);
	    XtSetSensitive(takemsg_w, True);
	}
}

/*ARGSUSED*/
static
void	invitationRepeater(client_data, id)
	XtPointer client_data;
	XtIntervalId *id;
{
	msgAnnouncer = False;
	XtRemoveTimeOut(requestId);
	XtSetSensitive(accept_w, False);
	XtSetSensitive(refuse_w, False);
	XtSetSensitive(putonhold_w, False);
	XtSetSensitive(takemsg_w, False);
	XtSetSensitive(typing_w, True);
	XtSetSensitive(dial_w, True);

	(void)xrelay_msg(&pc_msg, REFUSE_CALL);
	XtRemoveTimeOut(callIndicatorID);
	XtVaSetValues(topLevel, XtNiconPixmap, icon_pixmap, NULL);
}

/*ARGSUSED*/
static
void	ProcessRelayMsg(client_data, source, id)
	XtPointer client_data;
	int *source;
	XtInputId *id;
{
	LOCAL_MSG lmsg;
	Bool resetall, broadcast;
	static char buf[MAX_MSGBUF + 1];
	int repeatc, belltone, freqidx, ringc;

	extern int newMessages;
	extern Bool ismsgBoxUp;


	if ((read(*source, (char *)&lmsg, sizeof(lmsg))) != sizeof(lmsg)) {
	    issueError("  Experiencing some difficulties on the relay socket.", VSHORT_REPEAT, FQ5_INDX,
		       BELL_TONE5, 4);
	    return;
	}

	resetall = True;
	broadcast = True;
	freqidx = FQ4_INDX;
	belltone = BELL_TONE3;
	repeatc = SHORT_REPEAT;
	ringc = INFINITE_RING;

	switch(lmsg.type) {
	   case TALK_LINE: 
			   broadcast = False;
			   getTalkLine();
			   break;

	   case TALK_REQUEST: 
			   solidIcon = True;
			   XtVaSetValues(topLevel, XtNiconPixmap, iicon_pixmap, NULL);
	    		   callIndicatorID = XtAppAddTimeOut(xcomm, msgIntervals[FQ4_INDX], changeIcon, NULL);
			   issueRequest(&lmsg, buf, True);
			   resetall = False;
			   repeatc = MEDIUM_REPEAT;
			   freqidx = FQ5_INDX;
			   break;

	   case DAEMON_RESPONSE: 
			   ringc = 4;
			   freqidx = FQ5_INDX;
			   decodeDaemonResp(lmsg.id, buf);
			   break;

	   case LEFT_REQUEST:
			   ringc = 0;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "   Left request on %-s for %-s.", hisAddressBuf, hisNameBuf);
			   resetall = False;
			   break;

	   case RINGING_AGAIN:
			   ringc = 0;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "   Ringing %-s again.", hisNameBuf);
			   resetall = False;
			   break;

	   case CANCELED:
			   freqidx = FQ5_INDX;
			   ringc = 2;
			   (void)sprintf(buf, "   %-s hung up", hisNameBuf);
			   break;

	   case NUT_HOUSE:
			   ringc = 4;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Have U lost it all, man ?  U can't talk to yourself !!!");
			   break;

	   case TRYING_LOCAL:
			   ringc = 0;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "   Trying to contact local talk daemon.");
			   resetall = False;
			   break;

	   case TRYING_REMOTE:
			   ringc = 0;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "   Trying to contact talk daemon on %-s machine.", hisNameBuf);
			   resetall = False;
			   break;

	   case HASNT_CONNECTED:
			   freqidx = FQ6_INDX;
			   ringc = 0;
			   (void)sprintf(buf, "   talk daemon on %-s hasn't connected talk line yet.", hisAddressBuf);
			   resetall = False;
			   break;

	   case TRY_LATER:
			   freqidx = FQ6_INDX;
			   ringc = 2;
			   (void)sprintf(buf, "  Can't get hold of %-s, you might want to try again later",
					hisNameBuf);
			   break;

	   case FAILED:	   (void)sprintf(buf, "   The current task failed for some reason -- just one of those things.");
			   freqidx = FQ5_INDX;
			   ringc = 3;
			   break;

	   case LOST_CONNECTION:
			   ringc = 3;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "   Lost connection to %-s.", hisNameBuf);
			   break;

	   case NEW_MESSAGE:
			   if (!ismsgBoxUp) {
			       (void)sprintf(buf, "  You have new message%-c.",
				     newMessages ? 's':' ');
			       freqidx = FQ5_INDX;
			       ringc = 2;
			       XtVaSetValues(retrievemsg_w,
					    XtNlabel, "  Read  Msg  ",
					    NULL);
			   }
			   else {

			       broadcast = False;
			       newMessages++;
			   }
			   resetall = False;
			   break;

	   case HANGED_UP: removeFromMenu(lmsg.id, buf);
			   resetall = False;
			   XtRemoveTimeOut(callIndicatorID);
			   XtVaSetValues(topLevel, XtNiconPixmap, icon_pixmap, NULL);
			   ringc = 1;
			   freqidx = FQ5_INDX;
			   break;

	   case HIS_INFORMATION:
			   issueRequest(lmsg.line, buf, False);
			   broadcast = False;
			   resetall = False;
			   break;

	   case WILL_HOLD: addToMenu(lmsg.id);
			   broadcast = False;
			   break;

	   case MADE_CONTACT:
			   ringc = 0;
			   freqidx = FQ5_INDX;
			   resetall = False;
			   (void)sprintf(buf, "  Made contact with <%-s>.", hisAddressBuf);
			   break;
	   case MACHINE_UNKNOWN:
			   ringc = 2;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Couldn't get address for <%-s>.", hisAddressBuf);
			   break;

	   case RESTART:   (void)sprintf(buf, "  Sub-program might be out of whack -- recommand restart !!!");
			   freqidx = FQ5_INDX;
			   ringc = 8;
			   break;

	   case SEND_FILE_ERROR: 
			   ringc = 2;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Error while transferring talk line (socket), might want to restart.");
			   break;

	   case ON_HOLD:
			   ringc = 2;
			   freqidx = FQ6_INDX;
			   repeatc = MEDIUM_REPEAT;
			   (void)sprintf(buf, "  Connection established but call is temporarily on hold.");
			   XtSetSensitive(hangup_w, True);
			   resetall = False;
			   break;

	   case SELECT_ERROR:
			   ringc = 2;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Error while trying to detect(select) activity on socket.");
			   break;

	   case LISTEN_ERROR:
			   ringc = 2;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Error on attempt to listen on (%-s) line.", hisNameBuf);
			   break;

	   case ACCEPT_ERROR:
			   ringc = 2;
			   freqidx = FQ5_INDX;
			   (void)sprintf(buf, "  Error on attempt to accept connection.");
			   break;

	   default:	   (void)sprintf(buf, "  Error from sub-program: errno -- (%3d).", lmsg.type);
			   freqidx = FQ5_INDX;
			   ringc = 3;
	}
	if (resetall) {
	    clear_textwin(typing_w);
	    if (lmsg.type != NUT_HOUSE)
	        clear_textwin(status_w);
	    XtSetSensitive(dial_w, True);
	    XtSetSensitive(typing_w, True);
	    XtSetSensitive(hangup_w, False);
	    if (callwaiting > 0)
	        XtSetSensitive(pickcall_w, True);
	}
	if (broadcast)
	    issueError(buf, repeatc, freqidx, belltone, ringc);
}


int	decodeDaemonResp(id, buf)
	int id;
	char buf[];
{

	switch(id) {
	   case NOT_HERE:  (void)sprintf(buf, "  %-s is NOT logged on.", hisNameBuf);
			   break;
	   case FAILED:	   (void)sprintf(buf, "  (%-s) is too confused to respond to us.", hisAddressBuf);
			   break;
	   case MACHINE_UNKNOWN:
			   (void)sprintf(buf, "  (%-s) does not recognize us.", hisAddressBuf);
			   break;
	   case PERMISSION_DENIED:
			   (void)sprintf(buf, "  %-s is currently NOT accepting calls.", hisNameBuf);
			   break;
	   case UNKNOWN_REQUEST:
			   (void)sprintf(buf, "  (%-s) can only handle local talk request, not remote ones.",
				   hisAddressBuf);
			   break;
	   case BADVERSION:
			   (void)sprintf(buf, "  (%-s) indicates protocol mismatch. Incompatible versions.",
				   hisAddressBuf);
			   break;
	   case BADADDR:   (void)sprintf(buf, "  (%-s) indicates bad (addr).", hisAddressBuf);
			   break;
	   case BADCTLADDR:
			   (void)sprintf(buf, "  (%-s) indicates bad (ctl_addr).", hisAddressBuf);
			   break;
	   default:	   (void)sprintf(buf, "  Unknown response from daemon: %3d", id);
	}
}

